from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future.builtins import str
from future import standard_library
standard_library.install_hooks()
import re
from .tools import PROGRESSBAR, Queue, setup_progressbar


class GeneTreeNodeException(Exception):
    pass


class GeneTree(object):
    def __init__(self, root):
        self.root = root

    def __str__(self):
        return str(self.root) + ';'

    def __iter__(self):
        return iter(self.root)

    def write(self, NHX=True):
        return self.root.write(NHX) + ';'


class GeneTreeNode(object):

    events = {'speciation': ('S', 'F'),
              'duplication': ('D', 'T'),
              'loss': ('L', 'F'),
              'leaf': ('', 'F')}

    reg = re.compile(r'\W') # matches anything that's NOT a-z, A-Z, 0-9 or _

    def __init__(self, name, node_type, level=''):
        self.name            = name
        self.node_type       = node_type
        self.taxonomic_level = level
        self.children        = []
        self.parent          = None
        self.genes           = []
        self.fam_id          = ''

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def node_type(self):
        return self._node_type

    @node_type.setter
    def node_type(self, node_type):
        if not node_type in ['speciation', 'duplication', 'loss', 'leaf']:
            raise Exception('Unexpected node_type: {}'.format(node_type))
        self._node_type = node_type

    def add_child(self, child):
        self.children.append(child)
        child.parent = self

    def is_leaf(self):
        return len(self.children) == 0

    def is_inner(self):
        return not self.is_leaf()

    def __iter__(self):
        yield self
        for ch in self.children:
            for element in ch:
                yield element

    def __str__(self):
        return self.write(NHX=False)

    def build_NHX_string(self):
        NHX_string = '&&NHX'
        event, duplication = self.events[self.node_type]

        if self.node_type != 'leaf':
            NHX_string += ':Ev={0}'.format(event)
            if self.node_type != 'loss':
                NHX_string += ':D={0}'.format(duplication)
        if self.taxonomic_level > '':
            NHX_string += ':S={0}'.format(self.taxonomic_level)
        return '[' + NHX_string + ']'

    def delete(self, recursive=False):
        """ Deletes this node from the tree. If recursive=True it also
        deletes the entire subtree below this node. If recursive=False
        the subtree is attached to this node's parent.
        """
        if not self.parent:
            raise GeneTreeNodeException('You can\'t delete the root node')
        for child in self.children:
            self.parent.add_child(child)
            if recursive:
                child.delete(recursive)
        self.parent.children.remove(self)
        self.children = list()

    def _invert(self):
        """ Inverting its relationship with its parent (oedipal!)
        """
        self.add_child(self.parent)
        self.parent.children.remove(self)

    def reroot(self):
        """ Reroots the tree at this node
        """
        if not self.parent:
            return self

        stack = list()
        n = self
        while not (n.parent is None):
            stack.append(n)
            n = n.parent
        for node in stack[::-1]:
            node._invert()
        return self

    def write(self, NHX=True):
        name = self.name
        if self.reg.search(name):
            name = '"{0}"'.format(name)
        NHX_string = self.build_NHX_string()
        if self.is_leaf():
            return '{0}{1}'.format(name, (NHX_string if NHX else ''))
        else:
            subtree = ', '.join(child.write(NHX) for child in self.children)
            return '({0}){1}{2}'.format(subtree, name, (NHX_string if NHX else ''))

    def iter_preorder(self):
        """ Traverse the tree in preorder ordering: ancestors before
        descendents """
        yield self
        for child in self.children:
            for elem in child.iter_preorder():
                yield elem

    def iter_postorder(self):
        """ Traverse the tree in postorder ordering: descendents before
        ancestors """
        for child in self.children:
            for elem in child.iter_postorder():
                yield elem
        yield self

    def iter_levelorder(self):
        """ Traverse the tree in levelorder ordering: first the root, then
        nodes at distance==1 from the root, then distance==2, etc. """
        q = Queue()
        q.enqueue(self)
        while q:
            node = q.dequeue()
            yield node
            for child in node.children:
                q.enqueue(child)

    def iter_leaves(self):
        for elem in self.iter_preorder():
            if elem.is_leaf():
                yield elem

    def iter_inner_nodes(self):
        for elem in self.iter_preorder():
            if elem.is_inner():
                yield elem

class GeneTreeTracer(object):
    def __init__(self, parser, taxonomy):
        self.parser = parser
        self.tax = taxonomy
        self.trees = list()
        self._dup_counter = 0
        self._loss_counter = 0

    def _reset_counters(self):
        self._dup_counter = self._loss_counter = 0

    def _get_job_list(self, add_genelists=False):
        job_list = list()
        for n in self.tax:
            if (self.tax.root == n.name) or (n.up is None):
                for family in n.history.geneFamList:
                    job_list.append((n, family, None, add_genelists))
            else:
                for family in n.comparison.filter('novel'):
                    job_list.append((n, n.history[family.fam], None,
                        add_genelists))

        return sorted(job_list, key=lambda x: x[1])

    def trace_gene_families(self, add_genelists=False, verbosity=0):
        job_list = self._get_job_list(add_genelists)

        if PROGRESSBAR and verbosity > 0:
            pbar = setup_progressbar('Extracting {0} genetrees: '
                                     .format(len(job_list)), len(job_list))
            pbar.start()

        for i, job in enumerate(job_list, start=1):
            self.trace_gene_family(*job)
            self._reset_counters()
            if PROGRESSBAR and verbosity > 0:
                pbar.update(i)

        if PROGRESSBAR and verbosity > 0:
            pbar.finish()

    def _create_parent(self, name, node_type, level):
        parent = GeneTreeNode(name, node_type, level)
        self.trees.append(GeneTree(parent))
        return parent

    def trace_gene_family(self,
                          taxnode,
                          family,
                          parent=None,
                          add_genelists=False,
                          ):

        fam_id = family.getFamId()

        if taxnode.is_leaf():
            return

        for child in taxnode.down:
            comparison = child.comparison[fam_id]

            if comparison.event == 'lost':
                if parent is None:
                    parent = self._create_parent(taxnode.name, 'speciation',
                        taxnode.name)
                    if add_genelists:
                        parent.genes = family.getMemberGenes()
                    parent.fam_id = fam_id
                self._loss_counter += 1
                node = GeneTreeNode('LOSS_{0}_{1}'.format(self._loss_counter,
                                                          child.name),
                            node_type='loss',
                            level=child.name)
                node.fam_id = fam_id
                parent.add_child(node)

            elif comparison.event == 'identical':
                if parent is None:
                    parent = self._create_parent(taxnode.name, 'speciation',
                        taxnode.name)
                    if add_genelists:
                        parent.genes = family.getMemberGenes()
                    parent.fam_id = fam_id
                next_fam = child.history[fam_id]
                if child.is_leaf():
                    gene = next_fam.getMemberGenes()[0]
                    label = self.parser.mapGeneToXRef(gene)
                    node = GeneTreeNode(label, node_type='leaf',
                        level=child.name)
                else:
                    node = GeneTreeNode(child.name, node_type='speciation',
                        level=child.name)
                if add_genelists:
                    node.genes = next_fam.getMemberGenes()
                node.fam_id = fam_id
                parent.add_child(node)
                self.trace_gene_family(child, next_fam, node, add_genelists)

            elif comparison.event == 'duplicated':
                if parent is None:
                    parent = self._create_parent(taxnode.name, 'duplication',
                        taxnode.name)
                    if add_genelists:
                        parent.genes = family.getMemberGenes()
                    parent.fam_id = fam_id
                self._dup_counter += 1
                dup = GeneTreeNode('DUP_{0}'.format(self._dup_counter),
                    node_type='duplication', level='')
                dup.fam_id = fam_id
                parent.add_child(dup)
                if add_genelists:
                    dup.genes = list()
                for dup_id in comparison.into.split('; '):
                    next_fam = child.history[dup_id]
                    if child.is_leaf():
                        gene = next_fam.getMemberGenes()[0]
                        label = self.parser.mapGeneToXRef(gene)
                        node = GeneTreeNode(label, node_type='leaf', level=child.name)
                    else:
                        node = GeneTreeNode(child.name, node_type='speciation',
                            level=child.name)
                    if add_genelists:
                        node.genes = next_fam.getMemberGenes()
                        dup.genes.extend(node.genes)
                    node.fam_id = dup_id
                    dup.add_child(node)
                    self.trace_gene_family(child, next_fam, node, add_genelists)

        return parent



import os
import shlex
import sys
import threading
from abc import ABCMeta, abstractmethod, abstractproperty
from locale import getpreferredencoding
from subprocess import PIPE, Popen
import logging


IS_PY3 = sys.version_info[0] == 3
POSIX = 'posix' in sys.builtin_module_names
DEFAULT_ENCODING = getpreferredencoding() or "UTF-8"
logger = logging.getLogger(__name__)

if IS_PY3:
    from queue import Queue
else:
    from Queue import Queue


class ExternalProcessError(Exception):
    pass


def _py2_and_3_joiner(sep, joinable):
    """
    Allow '\n'.join(...) statements to work in Py2 and Py3.
    :param sep:
    :param joinable:
    :return:
    """
    if IS_PY3:
        sep = bytes(sep, DEFAULT_ENCODING)
    joined = sep.join(joinable)
    return joined.decode(DEFAULT_ENCODING) if IS_PY3 else joined


def _kwargs_to_args(kwargs, num_hyphens):
    processed = []
    for k, v in kwargs.items():
        # we're passing a short arg as a kwarg, example:
        # cut(d="\t")
        if v is not False:
            if len(k) == 1:
                processed.append('-' + k)
            else:
                processed.append('-' * num_hyphens + k)
            if v is not True:
                processed.append(v)
    return ' '.join(processed)


class AbstractCLI(object):
    """
    Abstract Base Class:
    Run an external program as a subprocess with non-blocking collection of stdout.
    This class uses subprocess.Popen to handle the system call, and threading.Thread
    to allow non-blocking reads of stderr and stdout.

    Specific program wrappers inherit from this. They should implement two methods:
    1: _default_exe and 2: _set_help.

    :Example:

    ::

       class SomeProgram(AbstractInternal):
           @property
           def _default_exe(self):
               return 'some_program'           # The usual name of the program binary
                                               # or alternatively specify a list of names

           def _set_help(self):
               self('--help', wait=True)       # calls 'some_program --help'
               self._help = self.get_stdout()  # puts the help output into self._help

           #### Possible Extra Requirement ####
           @property
           def _hyphen_policy(self):
               return 1       # If some_program arguments take a single leading hyphen
                              # then set this to 1 (i.e. to the number of leading hyphens)
                              # This only affects calling some_program using named arguments

    That's it!

    :seealso:

        See blog post: `<http://www.zultron.com/2012/06/python-subprocess-example-running-a-background-subprocess-with-non-blocking-output-processing/>`_
        Also, StackOverflow: `<https://stackoverflow.com/questions/375427/non-blocking-read-on-a-subprocess-pipe-in-python>`_

    Keyword handling borrowed from sh: `<https://amoffat.github.io/sh>`_

    """
    __metaclass__ = ABCMeta

    def __init__(self, executable=None, verbose=True):
        """
        Sets up the wrapper. A custom executable can be passed in, otherwise
        it will search the PATH.

        :param executable: Path to a custom executable to use
        :return:
        """
        exe = None
        if executable:
            exe = self._search_for_executable(executable)
        if exe is None:
            if isinstance(self._default_exe, list): 
                for default_name in self._default_exe:
                    exe = self._search_for_executable(default_name)
                    if exe is not None:
                        break
            else:
                exe = self._search_for_executable(self._default_exe)
            if not exe:
                logger.error("executable not found. PATH='{}'".format(os.getenv('PATH', '')))
                raise IOError(executable if executable else self._default_exe)

        self.exe = exe  # The wrapped executable
        self.verbose = False  # Controls printing of output
        self.stdout_q = Queue()  # The output-reading threads pipe output into
        self.stderr_q = Queue()  # these queues.
        self.stdout_l = list()  # Queued output gets placed in these lists,
        self.stderr_l = list()  # making it available to the caller
        self.stdout_thread = None  # monitoring thread that captures stdout
        self.stderr_thread = None  # monitoring thread that captures stderr
        self.process = None  # This holds the running process once the wrapped program is called
        self._help = None  # A place to hold a help string

    def __repr__(self):
        return '{}(executable=\'{}\')'.format(self.__class__.__name__, self.exe)

    # Private
    @abstractproperty
    def _default_exe(self):
        pass

    @property
    def _hyphen_policy(self):
        """
        Returns 'n', where 'n' is the number of hyphens prepended to commandline flags.
        Used internally when constructing command line arguments from parameters
        passed to __call__ as keywords.
        :return: 2 (as in 2 hyphens, '--')
        """
        return 2

    def _set_help(self):
        pass

    def _log_thread(self, pipe, queue):
        """
        Start a thread logging output from pipe
        """

        # thread function to log subprocess output (LOG is a queue)
        def enqueue_output(out, q):
            for line in iter(out.readline, b''):
                q.put(line.rstrip())
            out.close()

        # start thread
        t = threading.Thread(target=enqueue_output,
                                  args=(pipe, queue))
        t.daemon = True  # thread dies with the program
        t.start()
        return t

    def _search_for_executable(self, executable):
        """
        Search for file give in "executable". If it is not found, we try the environment PATH.
        Returns either the absolute path to the found executable, or None if the executable
        couldn't be found.
        """
        # Check in case java wrapper is called where executable is called with "java -j XXX"
        if len(executable.split(" ")) > 1:
            return executable

        if os.path.isfile(executable):
            return os.path.abspath(executable)
        else:
            envpath = os.getenv('PATH')
            if envpath is None:
                return
            for path in envpath.split(os.pathsep):
                exe = os.path.join(path, executable)
                if os.path.isfile(exe):
                    return os.path.abspath(exe)

    # Public
    def __call__(self, cmd=None, wait=False, **flags):
        """
        Spawns the subprocess and the threads used to monitor stdout and stderr without blocking.
        :param cmd: Pass the command line arguments as a string
        :param wait: Block until the process returns
        :param flags: Pass the commandline arguments as a dictionary. Will be appended to any content in cmd.

        :return:
        """
        # Check there is not already a process running
        if self.running():
            self.kill()

        # Wipe any stdout/stderr from previous processes
        self.stderr_l = []
        self.stdout_l = []

        # Assemble command line
        if cmd is None:
            cmd = ''
        if flags:
            cmd = ' '.join([cmd.strip(), _kwargs_to_args(flags, self._hyphen_policy)])
        self.cmd = '{} {}'.format(self.exe, cmd)
        logger.debug('Running following command: {}'.format(self.cmd))

        # spawn
        self.process = Popen(shlex.split(self.cmd),
                             shell=False, stdout=PIPE, stderr=PIPE, bufsize=128, close_fds=POSIX)
        if self.verbose:
            print('Launched {} with PID {}'.format(self.exe, self.process.pid))

        # start stdout and stderr logging threads
        self.stdout_thread = self._log_thread(self.process.stdout, self.stdout_q)
        self.stderr_thread = self._log_thread(self.process.stderr, self.stderr_q)

        if wait:
            self.process.wait()
            self._join_threads()

    @property
    def help(self):
        """
        Returns a helpful string, preferably derived from the wrapped program
        :return:
        """
        if self._help is None:
            self._set_help()
        return self._help

    def _join_threads(self):
        for thread in [self.stdout_thread, self.stderr_thread]:
            if thread is not None and thread.is_alive():
                thread.join()
                thread = None

    def get_stderr(self, tail=None):
        """
        Returns current total output written to standard error.
        :param tail: Return this number of most-recent lines.
        :return: copy of stderr stream
        """
        while not self.stderr_q.empty():
            self.stderr_l.append(self.stderr_q.get_nowait())
        if tail is None:
            tail = len(self.stderr_l)
        return _py2_and_3_joiner('\n', self.stderr_l[:tail])

    def get_stdout(self, tail=None):
        """
        Returns current total output written to standard output.
        :param tail: Return this number of most-recent lines.
        :return: copy of stdout stream
        """
        while not self.stdout_q.empty():
            self.stdout_l.append(self.stdout_q.get_nowait())
        if tail is None:
            tail = len(self.stdout_l)
        return _py2_and_3_joiner('\n', self.stdout_l[:tail])

    def finished(self):
        """
        Check if the running process is finished. Raises an exception if no process has ever been launched.
        :return: bool
        """
        if self.process is None:
            raise ExternalProcessError('No process has been launched from this instance')
        return self.process.poll() is not None

    def kill(self):
        """
        Kill the running process (if there is one)
        :return: void
        """
        if self.running():
            if self.verbose:
                print('Killing {} with PID {}'.format(self.exe, self.process.pid))
            self.process.kill()

            # Threads *should* tidy up by themselves, but we do it explicitly
            self._join_threads()

    def running(self):
        """
        True if there is a running process. False if either no process is associated with this instance,
        or if the associated process has finished.
        :return: bool
        """
        if self.process is None:
            return False
        return self.process.poll() is None



from ..utils import auto_open
import collections
from time import time
import xml.etree.ElementTree as etree
import Bio.Phylo
from pathlib import Path
import logging
logger = logging.getLogger(__name__)


class TaxLevel:
    """TaxLevel class

    This class represents a clade of interest to extract the flat groups.
    You can instantiate it with just a clade parameter, in which case it
    must match exactly the <property name="TaxRange" value="XXX"> value
    of your level of interest.

    A better way to use this class is to pass clade and a phyloxml file
    with the clade definitions (a species tree) and select the clade as
    an internal name of this phylogeny. In this case, all the species
    of the orthoxml must match the extend species in this phylogeny.

    """
    def __init__(self, clade: str, fn_phyloxml=None):
        if fn_phyloxml is not None:
            with open(fn_phyloxml, 'rt') as fh:
                tree = Bio.Phylo.read(fh, 'phyloxml')
            target_clade = next(tree.find_clades(target=clade))
            self.species = frozenset([n.taxonomy.code for n in target_clade.get_terminals()])
            self.taxid = target_clade.taxonomy.id.value
        else:
            logger.warning("TaxLevel only initialized with a clade string. Will only match exact taxrange matches!")
            self.species = set([])
            self.taxid = ""
        self.name = clade

    def all_in_clade(self, genes):
        return set(g.species for g in genes).issubset(self.species)

    def does_match(self, name_or_id):
        return name_or_id == self.name or name_or_id == self.taxid


Gene = collections.namedtuple("Gene", "xref species internal_id")


def _callback_group(node):
    return node.get('id')


class GroupExtractor(object):
    def __init__(self, target_clade:TaxLevel, gene_attr="protId", callback=None):
        self.processed_stats = {'last': time(), 'processed_toplevel': 0}
        self.target_clade = target_clade
        self.gene_attr = gene_attr
        self.genes = {}
        self.callback = callback
        if callback == "group_id":
            self.callback = _callback_group

    def add_genome_genes(self, genome_node):
        genome_name = genome_node.get('name', None)
        if genome_name is None:
            genome_name = genome_node.get("NCBITaxId")

        generef_2_xref = {}
        for gene in genome_node.findall('.//{http://orthoXML.org/2011/}gene'):
            gene_id = gene.get('id')
            gene_prot_id = gene.get(self.gene_attr)
            generef_2_xref[gene_id] = Gene(gene_prot_id, genome_name, gene_id)
        self.genes.update(generef_2_xref)
        return True

    def _collect_genes(self, node):
        genes = set([]); to_rem = []
        for child in node.iter():
            if child == node:
                continue
            if child.tag == "{http://orthoXML.org/2011/}geneRef":
                try:
                    genes.add(self.genes[child.get('id')])
                except KeyError:
                    logger.info(f"ignoring gene(id={child.get('id')}), probably in skip set.")
                    pass
                to_rem.append(child)
            elif child.tag == "{http://orthoXML.org/2011/}orthologGroup":
                genes.update((n for n in child.text if isinstance(n, Gene)))
                to_rem.append(child)
        for c in to_rem:
            try:
                node.remove(c)
            except ValueError as e:
                # this is not a direct child of node. we ignore this potential
                # memory-leak as the entire group will be deleted at latest once
                # we reach the root orthologGroup node.
                pass
        return genes

    def merge_children(self, node):
        genes = self._collect_genes(node)
        node.text = genes

    def get_group(self, node):
        if self.callback is not None:
            return node.text, self.callback(node)
        return node.text

    def analyze_and_yield_groups(self, node, is_toplevel=False):
        genes = self._collect_genes(node)
        if self.target_clade.all_in_clade(genes):
            if is_toplevel:
                logger.debug("dumping toplevel hog with {} genes (hog_id: {})".format(len(genes), node.get('id')))
                if self.callback is not None:
                    yield genes, self.callback(node)
                else:
                    yield genes
            else:
                node.clear()
                node.text = genes
        else:
            for group in node.iter("{http://orthoXML.org/2011/}orthologGroup"):
                if group == node:
                    continue
                if self.target_clade.all_in_clade(group.text):
                    logger.debug("found hog with {} genes".format(len(group.text)))
                    yield self.get_group(group)
            node.clear()
            node.text = genes

    def handle_duplication_node(self, elem):
        pass


class MarkerGroupExtractor(GroupExtractor):
    def handle_duplication_node(self, elem):
        nr_children = [len(self._collect_genes(child)) for child in elem]
        max_pos = nr_children.index(max(nr_children))
        for i, child in enumerate(elem):
            if i != max_pos:
                elem.remove(child)


def parse_orthoxml(fh, processor:GroupExtractor):
    nsmap = {}
    og_level = 0
    extract_at_depth = -2
    if processor.target_clade == None:
        extract_at_depth = 0

    def fixtag(tag, ns=""):
        return "{" + nsmap[ns] + "}" + tag

    logger.info("start mapping of orthoxml formatted input file")
    for event, elem in etree.iterparse(fh, events=('start-ns', 'start', 'end')):
        if event == 'start-ns':
            ns, url = elem
            nsmap[ns] = url
        elif event == 'start' and elem.tag == fixtag('orthologGroup'):
            og_level += 1
        elif event == 'start' and elem.tag == fixtag('property') and processor.target_clade is not None:
            if og_level > 0 and elem.get('name').lower() in ("taxid", "taxrange", "taxonomic_range", "taxon_id", "ncbitaxid"):
                if processor.target_clade.does_match(elem.get('value')):
                    extract_at_depth = og_level-1
        elif event == 'end':
            if elem.tag == fixtag('orthologGroup'):
                og_level -= 1
                if extract_at_depth == -2:
                    # no level annotations, we need to check ourself if we are at the right level
                    yield from processor.analyze_and_yield_groups(elem, is_toplevel=(og_level == 0))
                elif extract_at_depth >= 0:
                    # we have taxonomic annotations. combine children and yield group in case
                    # we are at the right level
                    processor.merge_children(elem)
                    if og_level == extract_at_depth:
                        logger.debug("dumping annotated group with {} genes".format(len(elem.text)))
                        yield processor.get_group(elem)
                        elem.clear()
                        extract_at_depth = -1 if processor.target_clade is not None else 0
                if og_level == 0:
                    elem.clear()
            elif elem.tag == fixtag('paralogGroup'):
                processor.handle_duplication_node(elem)
            elif elem.tag == fixtag('species'):
                processor.add_genome_genes(elem)
                elem.clear()


def extract_group_at_level(f, processor):
    if isinstance(f, (str, bytes, Path)):
        with auto_open(f, 'rt') as fh:
            yield from parse_orthoxml(fh, processor)
    else:
        yield from parse_orthoxml(f, processor)


def extract_flat_groups_at_level(f, protein_attribute="protId", level=None, callback=None):
    """Iterates over the groups defined in an orthoxml file at a specific taxonomic level.

    This function yields groups of Genes defined in an orthoxml file for a specific
    taxonomic level. You could for example extract groups at the level of Mammalia, that
    had a duplication event prior to the last common ancestor of the mammalia. For that,
    one needs to pass in a `class:TaxLevel` instance for Mammalia. If the level parameter
    is None (default), all the roothogs are returned.

    :Note:
    If you instantiate the TaxLevel class with only a name, that name needs to match
    the <property name="TaxLevel" value="xxx"> exactly. also, if a taxonomic
    level is not annotated, the parser will *NOT* return the group at this level.
    It is much safer to instantiate the TaxLevel instance with a PhyloXML file and the
    associated clade name - in this case any group that contains only proteins of that
    clade will be returned.

    :Example:
    >>> toplevel_groups = []
    >>> oxml = "tests/coverage_test_files/simpleEx.orthoxml"
    >>> for grp, grp_id in extract_flat_groups_at_level(oxml, callback="group_id"):
    ...     toplevel_groups.append((set(g.xref for g in grp), grp_id))
    [({"XENTR1", "CANFA1", "HUMAN1", "PANTR1", "MOUSE1", "RATNO1"}, "1"),
     ({"HUMAN2", "PANTR2", "CANFA2", "MOUSE2"}, "2"),
     ({"XENTR3", "MOUSE4", "MOUSE3", "CANFA3", "PANTR4", "PANTR3", "HUMAN3"}, "3")]


    :param f: filehandle of filename containing the orthoxml file.

    :param protein_attribute: the protein_attribute that should be read from the orthoxml
                              and that should be returned as xref attribute of the Gene.
                              The default value is the `protId` attribute.

    :param level: a `TaxLevel` instance.

    :param callback: a callback function that accepts the node element of the group that
                     contains all the members of the desired clade.
                     As callback, the user can also pass the string `group_id` in which
                     case the group_id will be returned.

    :returns: If no callback is provided, the method yields lists of Gene elements. Otherwise
              it yields tuples with (List[Gene], return_value_of_callback)

    """
    if level is not None:
        if not isinstance(level, TaxLevel):
            raise ValueError("level argument should be a TaxLevel instance.")
    processor = GroupExtractor(target_clade=level, gene_attr=protein_attribute, callback=callback)
    yield from extract_group_at_level(f, processor)


def extract_marker_groups_at_level(f, protein_attribute="protId", level=None, callback=None):
    """
    Iterates over the groups defined in an orthoxml file at a specific taxonomic level.

    This function yields flat groups of Genes defined in an orthoxml file for a specific
    taxonomic level, which are all orthologous to each other. This is done by always selecting
    the most heavy child group for paralogGroup nodes. You could for example extract groups
    at the level of Mammalia, that had a duplication event prior to the last common ancestor of the mammalia. For that,
    one needs to pass in a `class:TaxLevel` instance for Mammalia. If the level parameter
    is None (default), markers from all the roothogs are returned.

    :Note:
    If you instantiate the TaxLevel class with only a name, that name needs to match
    the <property name="TaxLevel" value="xxx"> exactly. also, if a taxonomic
    level is not annotated, the parser will *NOT* return the group at this level.
    It is much safer to instantiate the TaxLevel instance with a PhyloXML file and the
    associated clade name - in this case any group that contains only proteins of that
    clade will be returned.

    :Example:
    >>> markers = []
    >>> oxml = "tests/coverage_test_files/simpleEx.orthoxml"
    >>> for grp, grp_id in extract_marker_groups_at_level(oxml, callback="group_id"):
    ...     markers.append((set(g.xref for g in grp), grp_id))
    [({"XENTR1", "CANFA1", "HUMAN1", "PANTR1", "MOUSE1", "RATNO1"}, "1"),
     ({"HUMAN2", "PANTR2", "CANFA2", "MOUSE2"}, "2"),
     ({"XENTR3", "MOUSE3", "CANFA3", "PANTR3", "HUMAN3"}, "3")]


    :param f: filehandle of filename containing the orthoxml file.

    :param protein_attribute: the protein_attribute that should be read from the orthoxml
                              and that should be returned as xref attribute of the Gene.
                              The default value is the `protId` attribute.
    :param level: a `TaxLevel` instance.

    :param callback: a callback function that accepts the node element of the group that
                     contains all the members of the desired clade.
                     As callback, the user can also pass the string `group_id` in which
                     case the group_id will be returned.

    :returns: If no callback is provided, the method yields lists of Gene elements. Otherwise
              it yields tuples with (List[Gene], return_value_of_callback)
    """
    if level is not None:
        if not isinstance(level, TaxLevel):
            raise ValueError("level argument should be a TaxLevel instance.")
    processor = MarkerGroupExtractor(target_clade=level, gene_attr=protein_attribute, callback=callback)
    yield from extract_group_at_level(f, processor)


from xml.etree import ElementTree as ET
from typing import List, Iterable
from random import randint



class GeneRefManager:
    def __init__(self):
        self.xrefs = {}
        self.ids = set([])

    def _random_unused_id(self):
        while True:
            cand = randint(100000, 1000000000)
            if str(cand) not in self.ids:
                return str(cand)

    def register_and_reassign(self, gene_nodes:Iterable[ET.Element]):
        update_ids = {}
        to_rem = []
        for gene in gene_nodes:
            if gene.attrib['id'] in self.ids:
                if gene.attrib['protId'] in self.xrefs:
                    # protId already in set. is it unique? if yes, no action, otherwise error
                    if self.xrefs[gene.attrib['protId']] != gene.attrib['id']:
                        raise ValueError("protId '{}' is used several times with different gene id :'{},'{}'"
                                         .format(gene.attrib['protId'], self.xrefs[gene.attrib['protId']], gene.attrib['id']))
                    else:
                        to_rem.append(gene.attrib['id'])
                        continue
                else:
                    # reassign internal gene id.
                    new_id = self._random_unused_id()
                    update_ids[gene.attrib['id']] = new_id
                    gene.attrib['id'] = new_id

            self.xrefs[gene.attrib['protId']] = gene.attrib['id']
            self.ids.add(gene.attrib['id'])
        return update_ids, to_rem


class Merger:
    def __init__(self, first):
        self.NS = "http://orthoXML.org/2011/"
        ET.register_namespace("", self.NS)
        self.doc = ET.parse(first)
        self.root = self.doc.getroot()

        self.all_species = set(z.attrib['name'] for z in self.doc.findall('./{{{}}}species'.format(self.NS)))
        self.all_genes = GeneRefManager()
        self.all_genes.register_and_reassign(
            self.doc.findall("./{{{0}}}species/{{{0}}}database/{{{0}}}genes/{{{0}}}gene".format(self.NS))
        )

    def merge_file(self, other):
        gene_id_updates, to_rem = self.all_genes.register_and_reassign(
            other.findall("./{{{0}}}species/{{{0}}}database/{{{0}}}genes/{{{0}}}gene".format(self.NS)))
        self._remove_unnecessary_genes(other, to_rem)
        self._update_geneRef_ids(other.find('./{{{}}}groups'.format(self.NS)), gene_id_updates)

        for sp in other.findall("./{{{}}}species".format(self.NS)):
            if sp.attrib['name'] not in self.all_species:
                species_seen = False
                for i, el in enumerate(self.root):
                    if el.tag == "{{{}}}species".format(self.NS):
                        species_seen = True
                    elif species_seen:
                        break
                self.root.insert(i, sp)
                self.all_species.add(sp.attrib['name'])
            else:
                db = self.root.find("./{{{0}}}species[@name='{1}']/{{{0}}}database/{{{0}}}genes".format(self.NS, sp.attrib['name']))
                for g in sp.iterfind(".//{{{}}}gene".format(self.NS)):
                    db.append(g)
        grps = self.root.find("./{{{}}}groups".format(self.NS))
        for g in other.find("./{{{}}}groups".format(self.NS)):
            grps.append(g)

    def _update_geneRef_ids(self, root, gene_id_updates):
        for old_id, new_id in gene_id_updates.items():
            for g in root.iterfind(".//{{{0}}}geneRef[@id='{1}']".format(self.NS, old_id)):
                g.attrib['id'] = new_id

    def _remove_unnecessary_genes(self, root, to_rem):
        for e in to_rem:
            parent = root.find("./{{{0}}}species/{{{0}}}database/{{{0}}}genes/{{{0}}}gene[@id='{1}']/.."
                               .format(self.NS, e))
            child = parent.find("./{{{0}}}gene[@id='{1}']".format(self.NS, e))
            parent.remove(child)




    def write(self, fh):
        self.doc.write(fh, xml_declaration=True, encoding="UTF-8", default_namespace=None)


def merge_orthoxml_files(out, files):
    """function to merge several orthoxml files into a single orthoxml file that contains all groups.

    This function combines several orthoxml files into a single orthoxml file that
    contains all the groups and maintains a valid definition block of the species
    and their genes. The protId attributes among all the orthoxml files need to be
    either unique or being at least assigned to the same internal gene id; in that
    case it is assumed that it is the same gene across the different files and it
    can be merged.
    if the gene id attribute is the same two or more orthoxml files, but their
    protId value is different, a new gene id value is generated and the geneRef
    values are updated accordingly.

    :param out: a path or a filehandle object where the combined orthoxml data should
                be written to.

    :param files: a list of paths or filehandle objects (of valid orthoxml format) that
                should be merged.

    """

    first = files.pop()
    merger = Merger(first)
    for f in files:
        merger.merge_file(ET.parse(f).getroot())

    return merger.write(out)




from ..utils import auto_open
# import collections
# from time import time
from lxml import etree as ET
# import Bio.Phylo
from typing import Iterable
from pathlib import Path
import logging
logger = logging.getLogger(__name__)

class HOGFilter:
    def __init__(self, score:str, value:float):
        self.score = score
        self.value = value

    def remove(self, score_id, value):
        return score_id == self.score and self.value > float(value)


class OrthoXMLFilterProcesser:

    def __init__(self, filters:Iterable[HOGFilter]=None):
        self.filters = list(filters)

    def add_filter(self, filter:HOGFilter):
        self.filters.append(filter)

    def process(self, fh):
        NS = "http://orthoXML.org/2011/"
        self.doc = ET.parse(fh)
        root = self.doc.getroot()
        to_rem = []
        for hog in root.iterfind('.//{{{0}}}orthologGroup'.format(NS)):
            score = hog.find('./{{{0}}}score'.format(NS))
            if score is None:
                continue
            for filt in self.filters:
                if filt.remove(score.get('id'), score.get('value')):
                    to_rem.append(hog)
                    break
        logger.info(f"will remove {len(to_rem)} hogs")
        for h in to_rem:
            parent = h.getparent()
            if 'id' in h.attrib:
                logger.info("removing hog " + str(h) + " line " + str(h.sourceline) + " " +str(h.attrib['id']))
            else:
                logger.info("removing hog " + str(h) + " line " + str(h.sourceline))
            if parent:
                parent.remove(h)
                if sum(c.tag == "{{{0}}}orthologGroup".format(NS) for c in parent) == 0:
                    if 'id' in parent.attrib:
                        logger.info("consider deleting the empty parent hog "+str(parent)+" line "+str(parent.sourceline)+" "+str(parent.attrib['id']))
                    else:
                        logger.info("consider deleting the empty parent hog " + str(parent) + " line "+str(parent.sourceline))
                    to_rem.append(parent)

    def write(self, fh):
        self.doc.write(fh, xml_declaration=True, encoding="UTF-8")



def filter_orthoxml_file(source_orthoxml, out, filter: HOGFilter):
    processor = OrthoXMLFilterProcesser([filter])
    if isinstance(source_orthoxml, (str, bytes, Path)):
        with auto_open(source_orthoxml, 'rt') as fh:
            processor.process(fh)
    else:
        processor.process(source_orthoxml)
    processor.write(out)







import os.path
from shutil import which

from . import _utils_roothog
from ._wrappers import logger
from . import __version__ as fastoma_version



"""

fastoma-infer-roothogs --proteomes proteome --hogmap hogmap --out-rhog-folder omamer_rhogs -vv

"""


def fastoma_infer_roothogs():
    import argparse
    parser = argparse.ArgumentParser(description="checking parameters for FastOMA")
    parser.add_argument("--version", action="version", version="FastOMA v"+fastoma_version)
    parser.add_argument("--proteomes", required=True, help="Path to the folder containing the input proteomes")
    parser.add_argument("--splice", help="Path to the folder containing the splice information files")
    parser.add_argument("--hogmap", help="Path to the folder containing the hogmap files")
    parser.add_argument("--out-rhog-folder", required=True, help="Folder where the roothog fasta files are written") #out_rhog_folder
    parser.add_argument('-v', action="count", default=0, help="Increase verbosity to info/debug")

    parser.add_argument("--mergHOG-ratioMax-thresh", required=False, type=float, default=0.8, help="For merging rootHOGs, threshold of ratioMax ") # mergHOG_ratioMax_thresh
    parser.add_argument("--mergHOG-ratioMin-thresh", required=False, type=float, default=0.9, help="For merging rootHOGs, threshold of ratioMin ") # mergHOG_ratioMin_thresh
    parser.add_argument("--mergHOG-shared-thresh", required=False, type=float, default=10, help="For merging rootHOGs, threshold of number shared proteins ") # mergHOG_shared_thresh
    parser.add_argument("--mergHOG-fscore-thresh", required=False, type=float, default=70, help="For merging rootHOGs, threshold of famlut score shared proteins ") # mergHOG_fscore_thresh
    parser.add_argument("--big-rhog-size", required=False, type=int, default=50*1000, help= "For big rootHOGs, we have different heuristics") # big_rhog_size
    parser.add_argument("--big-fscore-thresh", required=False, type=int, default=95, help="For huge rootHOGs, we have different heuristics, like filtering low family score protiens") # big_fscore_thresh

    conf_infer_roothogs = parser.parse_args()
    logger.setLevel(level=30 - 10 * min(conf_infer_roothogs.v, 2))
    logger.debug("Arguments: %s", conf_infer_roothogs)

    species_names, prot_recs_lists, fasta_format_keep = _utils_roothog.parse_proteomes(conf_infer_roothogs.proteomes)  # optional input folder
    prot_recs_all = _utils_roothog.add_species_name_prot_id( prot_recs_lists)

    hogmaps, unmapped = _utils_roothog.parse_hogmap_omamer(species_names, fasta_format_keep, folder= conf_infer_roothogs.hogmap)  # optional input folder

    splice_files = conf_infer_roothogs.splice is not None and os.path.exists(conf_infer_roothogs.splice)
    if splice_files:
        isoform_by_gene_all = _utils_roothog.parse_isoform_file(species_names, folder=conf_infer_roothogs.splice)
        isoform_selected,  isoform_not_selected = _utils_roothog.find_nonbest_isoform(
            species_names, isoform_by_gene_all, hogmaps
        )
        _utils_roothog.write_isoform_selected(isoform_by_gene_all, isoform_selected, prot_recs_lists)
        # for each isoform file, there will be a file ending with _selected_isoforms.tsv
        hogmaps = _utils_roothog.handle_splice(hogmaps, isoform_not_selected)

    rhogs_prots = _utils_roothog.group_prots_roothogs(hogmaps)
    rhogs_prots = _utils_roothog.handle_singleton(rhogs_prots, hogmaps, conf_infer_roothogs)
    rhogs_prots = _utils_roothog.merge_rhogs2(hogmaps, rhogs_prots, conf_infer_roothogs)
    rhogs_prots = _utils_roothog.filter_big_roothogs(hogmaps, rhogs_prots, conf_infer_roothogs)

    min_rhog_size = 2
    rhogid_written_list = _utils_roothog.write_rhog(rhogs_prots, prot_recs_all, conf_infer_roothogs.out_rhog_folder, min_rhog_size)
    linclust_available=which("mmseqs")  # True #
    # if memseqs is not installed the output will be empty / None
    if linclust_available:
        num_unmapped_singleton = _utils_roothog.collect_unmapped_singleton(rhogs_prots, unmapped, prot_recs_all,  "singleton_unmapped.fa")
        if num_unmapped_singleton:
            result_linclust = _utils_roothog.run_linclust(fasta_to_cluster="singleton_unmapped.fa")
            logger.debug(" linclust is done %s", result_linclust)
            num_clusters = _utils_roothog.write_clusters(conf_infer_roothogs.out_rhog_folder, min_rhog_size)
            logger.debug("we wrote %d new clusters with linclust ", num_clusters)




import itertools
from abc import ABCMeta, abstractmethod
from enum import Enum
from Bio import AlignIO, SeqIO


from ...seq_utils import is_dna, identify_input, AlignmentInput
from .. import WrapperError




DataType = Enum('DataType', 'DNA PROTEIN UNKNOWN')


class Aligner(object):
    """
    Base class for wrappers of Multiple Sequence Aligner software

    The wrapper is written as a callable class.
    This can hold data (state) to do with the operation it performs, so it can keep results,
    execution times and other metadata, as well as perform the task.

    This is a base implementation to be extended. The important parts are
    __init__ (does the setup) and __call__ (does the work). All
    else are helper methods.

    :Example:

    ::

        callable_wrapper = ConcreteAligner(aln)
        result = callable_wrapper()
        time_taken = callable_wrapper.elapsed_time
        result_again = callable_wrapper.result

    """
    __metaclass__ = ABCMeta

    def __init__(self, input_, datatype=DataType.UNKNOWN, binary=None):
        """
        Should work the same whether you're working with a Biopython object or a file
        but the implementation differs, e.g. a Biopython object will need
        to be written temporarily to disk for the Aligner to work on it.

        :param input_: can be either a filename or a biopython multiple
            sequence alignment (a collection of :class:`Bio.SeqRecord.SeqRecord`)

        :param binary: is the alignment's executable file, or None. If set to
            None, it is assumed to be found in the PATH.

        :param datatype: means is it DNA or protein?
        """
        self.input_type = identify_input(input_)  # Figure out what it is - file or object

        if isinstance(datatype, str):
            try:
                datatype = getattr(DataType, datatype.upper())
            except AttributeError:
                raise ValueError("\"{}\" is an invalid datatype for an Aligner".format(datatype))
        if datatype == DataType.UNKNOWN:
            self.datatype = guess_datatype(input_, from_filename=self.input_type == AlignmentInput.FILENAME)
            if self.input_type == AlignmentInput.OBJECT:
                dup, input_ = itertools.tee(input_)
                self.datatype = guess_datatype(dup, False)
            else:
                self.datatype = guess_datatype(input_, True)
        else:
            self.datatype = datatype

        self.input = input_  # store it
        self.elapsed_time = None
        self.stdout = None
        self.stderr = None
        try:
            self.cli = self._init_cli(binary)
        except IOError as err:
            raise WrapperError('Error searching for binary: {}'.format(err))
        # End setup

    @abstractmethod
    def __call__(self, *args, **kwargs):
        """
        How to call the underlying aligner
        """
        pass

    @abstractmethod
    def _init_cli(self, binary):
        pass

import logging
logger = logging.getLogger()


def guess_datatype(alignment, from_filename=False):
    logger.warning("Guessing is not recommended - specify the sequence type with option datatype={DNA, PROTEIN}, be more confident")
    if from_filename:
        try:
            alignment = SeqIO.parse(alignment, 'fasta')
        except:
            alignment = SeqIO.parse(alignment, 'phylip-relaxed')
    return DataType.DNA if is_dna(alignment) else DataType.PROTEIN


# TODO: Break the identify_input function into two parts - one to work out the datatype, one to work out whether
# this is a file or an object



import collections
import csv
import itertools
from copy import deepcopy
from pathlib import Path
import gzip

import dendropy
from Bio import SeqIO
from Bio.Align import MultipleSeqAlignment
import concurrent.futures
import time
import os
import shutil
import pickle
import gc
import random
from ete3 import Tree, TreeNode
import xml.etree.ElementTree as ET
from typing import List

# import networkx as nx
# import matplotlib.pyplot as plt
from . import _wrappers
from . import _utils_subhog
from . import _utils_frag_SO_detection
from ._hog_class import HOG, Representative, split_hog
from ._utils_subhog import MSAFilter, MSAFilterElbow, MSAFilterTrimAL

from ._wrappers import logger
from .zoo.utils import unique

low_so_detection = True # detection of proteins with low species overlap score in gene tree
fragment_detection = True  # this also need to be consistent in _hog_class.py
keep_subhog_each_pickle = False
inferhog_resume_subhog = True
inferhog_max_workers_num = 6 # how parallel to work in a species tree, (won't help near the root)
inferhog_min_hog_size_xml = 2     # by setting this as 1, pyham won't work on xml output.
orthoxml_v03 = True


def read_infer_xml_rhogs_batch(rhogid_batch_list, inferhog_concurrent_on, pickles_rhog_folder, pickles_subhog_folder_all, rhogs_fa_folder, conf_infer_subhhogs):
    """
    infer subHOGs for a list of rootHOGs
    """
    logger.debug("Inferring subHOGs for batch of %d rootHOGs started.", len(rhogid_batch_list))
    hogs_rhog_xml_len_batch = []
    for rhogid in rhogid_batch_list:
        hogs_rhogs_xml_len = read_infer_xml_rhog(rhogid, inferhog_concurrent_on, pickles_rhog_folder,  pickles_subhog_folder_all, rhogs_fa_folder, conf_infer_subhhogs)
        hogs_rhog_xml_len_batch.append(hogs_rhogs_xml_len)

    return hogs_rhog_xml_len_batch


def read_infer_xml_rhog(rhogid, inferhog_concurrent_on, pickles_rhog_folder,  pickles_subhog_folder_all, rhogs_fa_folder, conf_infer_subhhogs):
    """
    infer subHOGs for a  rootHOGs
    """

    pickles_subhog_folder = pickles_subhog_folder_all + "/rhog_" + rhogid + "/"
    if not os.path.exists(pickles_subhog_folder):
        os.makedirs(pickles_subhog_folder)

    # if (_config.gene_trees_write or _config.msa_write) and not os.path.exists("./genetrees"):
    #     os.makedirs("./genetrees")

    logger.info("\n" + "==" * 10 + "\n Start working on root hog: " + rhogid + ". \n")
    rhog_i_prot_address = rhogs_fa_folder + "/HOG_" + rhogid+ ".fa"
    rhog_i = list(SeqIO.parse(rhog_i_prot_address, "fasta"))
    logger.info("number of proteins in the rHOG is " + str(len(rhog_i)) + ".")
    # the file "species_tree_checked.nwk" is created by the check_input.py
    (species_tree) = _utils_subhog.read_species_tree(conf_infer_subhhogs.species_tree)

    (species_tree, species_names_rhog, prot_names_rhog) = _utils_subhog.prepare_species_tree(rhog_i, species_tree, rhogid)
    species_names_rhog = list(set(species_names_rhog))
    logger.info("Number of unique species in rHOG " + rhogid + " is " + str(len(species_names_rhog)) + ".")

    if inferhog_concurrent_on:  # for big HOG we use parallelization at the level taxonomic level using concurrent
        hogs_a_rhog_num = infer_hogs_concurrent(species_tree, rhogid, pickles_subhog_folder_all, rhogs_fa_folder, conf_infer_subhhogs)
    else:
        hogs_a_rhog_num = infer_hogs_for_rhog_levels_recursively(species_tree, rhogid, pickles_subhog_folder_all, rhogs_fa_folder, conf_infer_subhhogs)
    # Output value hogs_a_rhog_num  is an integer= length. We save the output as pickle file at each taxonomic level.

    #####  Now read the final pickle file for this rootHOG
    root_node_name = species_tree.name
    pickle_subhog_file = pickles_subhog_folder + str(root_node_name) + ".pickle"
    with open(pickle_subhog_file, 'rb') as handle:
        hogs_a_rhog = pickle.load(handle)

    if not keep_subhog_each_pickle:
        shutil.rmtree(pickles_subhog_folder)

    hogs_rhogs_xml = []
    for hog_i in hogs_a_rhog:
        if len(hog_i._members) >= inferhog_min_hog_size_xml:
            # could be improved   # hogs_a_rhog_xml = hog_i.to_orthoxml(**gene_id_name)
            hogs_a_rhog_xml_raw = hog_i.to_orthoxml()    # <generef  >      <paralg object >
            if orthoxml_v03 and 'paralogGroup' in str(hogs_a_rhog_xml_raw) :
                # in version v0.3 of orthoxml, there shouldn't be any paralogGroup at root level. Let's put them inside an orthogroup should be in
                hog_elemnt = ET.Element('orthologGroup', attrib={"id": str(hog_i._hogid)})
                property_element = ET.SubElement(hog_elemnt, "property", attrib={"name": "TaxRange", "value": str(hog_i._tax_now.name)})
                hog_elemnt.append(hogs_a_rhog_xml_raw)
                hogs_a_rhog_xml = hog_elemnt
            else:
                hogs_a_rhog_xml = hogs_a_rhog_xml_raw
            hogs_rhogs_xml.append(hogs_a_rhog_xml)
        else:
            logger.debug("we are not reporting due to fastoma singleton hog |*|  " + str(list(hog_i._members)[0]))
            if len(hog_i._members) > 1:
                logger.warning("issue 166312309 this is not a singleton "+str(hog_i._members))


    pickles_rhog_file = pickles_rhog_folder + '/file_' + rhogid + '.pickle'
    with open(pickles_rhog_file, 'wb') as handle:
        # dill_pickle.dump(hogs_rhogs_xml, handle, protocol=dill_pickle.HIGHEST_PROTOCOL)
        pickle.dump(hogs_rhogs_xml, handle, protocol=pickle.HIGHEST_PROTOCOL)
    logger.info("All subHOGs for the rootHOG %s as OrthoXML format is written in %s", rhogid, pickles_rhog_file)
    # to see orthoxml as string, you might need to do it for different idx
    # idx=0; from xml.dom import minidom; import xml.etree.ElementTree as ET; minidom.parseString(ET.tostring(hogs_rhogs_xml[idx])).toprettyxml(indent="   ")
    del hogs_a_rhog  # to be memory efficient
    gc.collect()
    hogs_rhogs_xml_len = len(hogs_rhogs_xml)
    return hogs_rhogs_xml_len


def infer_hogs_concurrent(species_tree, rhogid, pickles_subhog_folder_all, rhogs_fa_folder, conf_infer_subhhogs):
    """
    infer subHOGs for a rootHOG using multi-threading (in parallel) on different taxanomic levels of species tree
    """

    pending_futures = {}
    with concurrent.futures.ThreadPoolExecutor(max_workers= inferhog_max_workers_num) as executor:
        for node in species_tree.traverse(strategy="preorder"):
            node.dependencies_fulfilled = set()  # a set
            # node.infer_submitted = False
            if node.is_leaf():
                future_id = executor.submit(singletone_hog_, node, rhogid, pickles_subhog_folder_all, rhogs_fa_folder)
                # singletone_hog_(sub_species_tree, rhogid, pickles_subhog_folder_all, rhogs_fa_folder)
                # {<Future at 0x7f1b48d9afa0 state=finished raised TypeError>: 'KORCO_'}
                pending_futures[future_id] = node.name
                # node.infer_submitted = True
        # to keep, have a family
        while len(pending_futures) > 0:
            time.sleep(0.01)
            future_id_list = list(pending_futures.keys())
            for future_id in future_id_list:
                if future_id.done(): # done means finished, but may be unsecussful.

                    species_node_name = pending_futures[future_id]
                    logger.debug("checking for rootHOG id "+str(rhogid)+" future object is done for node " +str(species_node_name))
                    future_id.result()
                    #logger.debug("the result of future is  " + str(future_id.result()))

                    del pending_futures[future_id]
                    species_node = species_tree.search_nodes(name=species_node_name)[0]
                    parent_node = species_node.up
                    if not parent_node:  # we reach the root
                        # assert len(pending_futures) == 0, str(species_node_name)+" "+rhogid_
                        assert species_node.name == species_tree.name
                        break
                    parent_node.dependencies_fulfilled.add(species_node_name)  # a set

                    childrend_parent_nodes = set(node.name for node in parent_node.get_children())
                    if parent_node.dependencies_fulfilled == childrend_parent_nodes:
                        #  if not parent_node.infer_submitted:
                        future_id_parent = executor.submit(infer_hogs_this_level, parent_node, rhogid, pickles_subhog_folder_all, conf_infer_subhhogs)
                        # parent_node.infer_submitted = True
                        # future_id_parent= parent_node.name+"aaa"
                        pending_futures[future_id_parent] = parent_node.name
                        # for future_id:  del pending_futures[future_id] i need another dictionary the other way arround to removes this futures

    return len(pending_futures) + 1


def infer_hogs_for_rhog_levels_recursively(sub_species_tree, rhogid, pickles_subhog_folder_all, rhogs_fa_folder, conf_infer_subhhogs):
    """
    infer subHOGs for a rootHOG using recursive function to traverse species tree (different taxanomic levels)
    """

    if sub_species_tree.is_leaf():
        singletone_hog_out = singletone_hog_(sub_species_tree, rhogid, pickles_subhog_folder_all, rhogs_fa_folder)
        #  out 1 =  succesful
        return singletone_hog_out
    children_nodes = sub_species_tree.children

    for node_species_tree_child in children_nodes:
        hogs_chrdn = infer_hogs_for_rhog_levels_recursively(node_species_tree_child, rhogid, pickles_subhog_folder_all, rhogs_fa_folder, conf_infer_subhhogs)
        # hogs_chrdn should be 1 hogs_chrdn_list.extend(hogs_chrdn)
    infer_hogs_this_level_out = infer_hogs_this_level(sub_species_tree, rhogid, pickles_subhog_folder_all, conf_infer_subhhogs)
    # hogs_this_level_list should be one
    return infer_hogs_this_level_out


def singletone_hog_(node_species_tree, rhogid, pickles_subhog_folder_all, rhogs_fa_folder):
    """
    create subHOGs for leaves of tree (species level). Each protein is a subHOG.
    """

    node_species_name = node_species_tree.name  # there is only one species (for the one protein)
    this_level_node_name = node_species_name
    pickles_subhog_folder = pickles_subhog_folder_all + "/rhog_" + rhogid + "/"
    # logger.debug(" ** inferhog_resume_subhog is " + str(_config.inferhog_resume_subhog))
    if inferhog_resume_subhog:
        # logger.debug("inferhog_resume_subhog is " + str(_config.inferhog_resume_subhog) + " so, we are reading from pickles.")
        pickle_subhog_file = pickles_subhog_folder + str(this_level_node_name) + ".pickle"
        # open already calculated subhogs , but not completed till root in previous run
        if os.path.exists(pickle_subhog_file):
            if os.path.getsize(pickle_subhog_file) > 3:  # 3 bytes
                with open(pickle_subhog_file, 'rb') as handle:
                    # i don't even need to open this even
                    # is output of pickle.load(handle) is chlired or this level ?
                    # todo I think I don't need to read the pickle file
                    hogs_this_level_list = pickle.load(handle) #[object class HOG HOG:4027_sub1,len=1,taxono=PSETE]
                    if hogs_this_level_list:
                        logger.debug("Level " + str(this_level_node_name) + " with " + str(len(hogs_this_level_list)) + " hogs is read from pickle.")
                        return len(hogs_this_level_list)
                    else:
                        logger.debug(" Issue  1238510: the pickle file for single tone is empty "+ str(hogs_this_level_list)+" " + rhogid)

    # logger.debug("reading protien / singletone HOG of  " + str(this_level_node_name))
    hogs_this_level_list = []
    rhog_i_prot_address = rhogs_fa_folder +"/HOG_"+rhogid+".fa"
    with open(rhog_i_prot_address, 'rt') as fasta:
        species_seq_generator = (rec for rec in SeqIO.parse(fasta, format="fasta")
                                 if rec.id.split('||')[1] == node_species_name)
        for prot in species_seq_generator:
            hog_leaf = HOG(prot, node_species_tree, rhogid)  # node_species_tree.name
            hogs_this_level_list.append(hog_leaf)
    pickle_subhog_file = pickles_subhog_folder + str(this_level_node_name)+".pickle"
    with open(pickle_subhog_file, 'wb') as handle:
        pickle.dump(hogs_this_level_list, handle, protocol=pickle.HIGHEST_PROTOCOL)
    logger.debug("HOGs for  " + str(this_level_node_name)+" including "+str(len(hogs_this_level_list))+ " hogs is written in pickle file.")

    return len(hogs_this_level_list)


def read_children_hogs(node_species_tree, rhogid, pickles_subhog_folder_all):
    this_level_node_name = node_species_tree.name
    pickles_subhog_folder = pickles_subhog_folder_all + "/rhog_" + rhogid + "/"
    # pickle_subhog_file = pickles_subhog_folder + str(this_level_node_name) + ".pickle"
    # if _config.inferhog_resume_subhog:  # open already calculated subhogs , but may not be completed in previous run
    #     if os.path.exists(pickle_subhog_file):
    #         if os.path.getsize(pickle_subhog_file) > 3:  # bytes
    #             with open(pickle_subhog_file, 'rb') as handle:
    #                 # i don't even need to open this even
    #                 hogs_children_level_list = pickle.load(handle)
    #                 if hogs_children_level_list:
    #                     return hogs_children_level_list

    children_name = [child.name for child in node_species_tree.children]
    hogs_children_level_list = []
    for child_name in children_name:
        pickle_subhog_file = pickles_subhog_folder + str(child_name) + ".pickle"
        with open(pickle_subhog_file, 'rb') as handle:
            hogs_children_level_list.extend(pickle.load(handle))  # when there is an error somewhere else, probably with paralelization, for big rootHOG, the root reason of the error won't shown up. but you stope here at worst which is late.
            #  todo, check file exist, how to handle if not
    logger.debug("Finding hogs for rhogid: " + rhogid+ ", for taxonomic level:" + str(
        node_species_tree.name) + " for species sub-tree:\n  " + str(node_species_tree.write(format=1, format_root_node=True)) + "\n")
    return hogs_children_level_list


RepLookup = collections.namedtuple("RepLookup", ["hog", "representative"])
class LevelHOGProcessor:
    def __init__(self, node_species_tree:TreeNode, subhogs:List[HOG], rhogid:str, conf):
        self.node_species_tree = node_species_tree
        self._frozen_hogs_before = set([])
        self.subhogs = {}
        for hog in subhogs:
            if hog.active:
                self.subhogs[hog.hogid] = hog
            else:
                self._frozen_hogs_before.add(hog)
        self.rhogid = rhogid
        self.conf = conf
        self._base_msa_tree_filename = "HOG_"+rhogid+"_"+str(node_species_tree.name)
        self._outname_step = 0
        self._removed_rep = set([])
        self._rep_lookup = self._prepare_lookups()
        self._msa_filter = self._instantiate_msa_filter()

    def _instantiate_msa_filter(self):
        if self.conf.msa_filter_method == "col-row-threshold":
            return MSAFilter(self, self.conf)
        elif self.conf.msa_filter_method == "col-elbow-row-threshold":
            return MSAFilterElbow(self, self.conf)
        elif self.conf.msa_filter_method == "trimal":
            return MSAFilterTrimAL(self, self.conf)
        else:
            raise ValueError("Invalid msa_filter_method value")

    def get_name_of_output(self, is_msa=False, is_tree=False):
        if (is_msa and self.conf.msa_write) or (is_tree and self.conf.gene_trees_write):
            self._outname_step += 1
            return f"{self._base_msa_tree_filename}_it{self._outname_step}"
        return None

    def _prepare_lookups(self):
        lookup = {}
        for hog in self.subhogs.values():
            for rep in hog.get_representatives():
                if rep.get_id() not in self._removed_rep:
                    lookup[rep.get_id()] = RepLookup(hog, rep)
        return lookup

    def find_most_divergent_representatives_from_genetree(self, genetree_subtree:TreeNode):
        """this method returns the N most divergent representatives from a rooted tree. within each cluster
        it selects the one representative sequence that corresponds to the the median distance from the root
        among the sequnces in the cluster.

        :param genetree: the rooted tree (with distances) that should be processed
        :return: the N most divergent representatives for that tree. N is taken from self.conf"""
        genetree = genetree_subtree.copy()
        # we remove the non-enabled representatives from the tree
        keep = [n for n in genetree.iter_leaves() if self._rep_lookup[n.name].representative.enabled]
        if len(keep) < len(genetree):
            genetree.prune(keep, preserve_branch_length=True)
        if len(genetree) <= self.conf.number_of_samples_per_hog:
            return [Representative(self._rep_lookup[n.name].representative) for n in genetree.iter_leaves()]

        # first, annotate the distance from the root
        for n in genetree.traverse():
            n.add_feature('dist_from_root', n.dist + (n.up.dist_from_root if n.up is not None and hasattr(n.up, 'dist_from_root') else 0))
        # now, let's get a list of all the internal nodes ordered with the dist_from_root and find which distance
        # results in number_of_samples_per_hog subtrees that are most divergent
        ordered = sorted([n for n in genetree.traverse() if not n.is_leaf()], key=lambda x: x.dist_from_root)
        idx = min(len(ordered), self.conf.number_of_samples_per_hog) - 1  # in case of multi-fucations, len(ordered) can be shorter than number_of_samples_per_hog
        cutoff = ordered[idx].dist_from_root
        sub_clades = [sub for sub in genetree.iter_leaves(
            is_leaf_fn=lambda n: len(n.children) == 0 or n.dist_from_root >= cutoff
        )]
        # now, let's select for each sub_clade the one leave that is close to the median dist_from_root (no outliers)
        representatives = []
        for sub in sub_clades:
            order_by_dist = sorted((n for n in sub.iter_leaves()), key=lambda n: n.dist_from_root)
            median_elem = order_by_dist[len(order_by_dist)//2]
            new_rep = Representative(self._rep_lookup[median_elem.name].representative,
                                     [self._rep_lookup[z.name].representative for z in order_by_dist])
            representatives.append(new_rep)
        return representatives

    def _get_merge_candidates_with_hogids(self, reconciled_genetree:TreeNode):
        for top_speciation_node in reconciled_genetree.iter_leaves(
                is_leaf_fn=lambda n: len(n.children) == 0 or (hasattr(n, 'evoltype') and n.evoltype == "S")
        ):
            hogids_in_subtree = set([])
            for n in top_speciation_node.iter_leaves():
                n.add_feature('hogid', self._rep_lookup[n.name].hog.hogid)
                hogids_in_subtree.add(n.hogid)
            top_speciation_node.add_feature('hogids', hogids_in_subtree)
            yield top_speciation_node

    def write_msa_or_tree_if_necessary(self, elem, fn_suffix="", features=[]):
        is_tree = isinstance(elem, TreeNode)
        is_msa = isinstance(elem, MultipleSeqAlignment)
        fn = self.get_name_of_output(is_tree=is_tree, is_msa=is_msa)
        if fn is not None:
            fn = Path(fn + fn_suffix)
            with open(fn, 'wt') as fout:
                if is_tree:
                    fout.write(elem.write(format=1, features=features, format_root_node=True))
                elif is_msa:
                    SeqIO.write(elem, fout, format="fasta")
            if is_tree:
                fn2 = fn.stem + ".tsv"
                with open(fn2, 'wt') as fout:
                    writer = csv.writer(fout, dialect="excel-tab")
                    writer.writerow(["id", "hogid", "hog_size", "representative_size", "enabled"])
                    for n in elem.iter_leaves():
                        rep_val = self._rep_lookup[n.name]
                        writer.writerow([n.name, rep_val.hog.hogid, len(rep_val.hog.get_members()), len(rep_val.representative.get_subelements()), rep_val.representative.enabled])
                fn3 = fn.stem + "-rep.tsv.gz"
                with gzip.open(fn3, 'wt') as fout:
                    writer = csv.writer(fout, dialect="excel-tab")
                    writer.writerow(["id", "representative"])
                    for n in elem.iter_leaves():
                        rep_val = self._rep_lookup[n.name]
                        for member in rep_val.representative.get_subelements():
                            writer.writerow([n.name, member])

    def align_subhogs(self):
        sub_msas = [hog.get_msa() for hog in self.subhogs.values() if len(hog.get_msa()) > 0]
        logger.debug(f"Merging {len(sub_msas)} MSAs for rhog: {self.rhogid}, level: {self.node_species_tree.name}")
        if len(sub_msas) == 0:
            logger.info(
                f"Issue 1455, merged_msa is empty {self.rhogid}, for taxonomic level: {self.node_species_tree.name}")
            raise ValueError("Issue 1455, merged_msa is empty")
        if len(sub_msas) > 1:
            merged_msa = _wrappers.merge_msa(sub_msas)
            self.write_msa_or_tree_if_necessary(merged_msa, fn_suffix=".fa")

            if fragment_detection:
                prot_dubious_msa_list, seq_dubious_msa_list = _utils_frag_SO_detection.find_prot_dubious_msa(
                    merged_msa, self.conf)
                if len(prot_dubious_msa_list) > 0:
                    logger.info(f"found fragments: {prot_dubious_msa_list} with {seq_dubious_msa_list}")
                    logger.error("Handling of fragments not yet implemented. Ignoring for now")
        else:
            merged_msa = sub_msas[0]   # when only on  child, the rest msa is empty.
        logger.debug(f"All sub-hogs are merged, merged_msa ({len(merged_msa)},{len(merged_msa[0])}) for "
                     f"rhog: {self.rhogid}, taxonomic level: {self.node_species_tree.name}")
        return merged_msa

    def filter_msa(self, msa):
        if self._msa_filter is None:
            self._msa_filter = MSAFilter(self, self.conf)
        filtered_msa, removed_ids = self._msa_filter.filter_msa(msa)
        if len(removed_ids) > 0:
            self._remove_representatives(ids=removed_ids)
        return filtered_msa

    def _remove_representatives(self, ids):
        self._removed_rep.update(ids)
        hogids_to_remove = []
        for hog_id, hog in self.subhogs.items():
            for rep in hog.get_representatives():
                if rep.get_id() not in self._removed_rep:
                    break
            else:
                # going here if not a break -> hog disappeared. remove it from current level and freeze it at level below
                hogids_to_remove.append(hog_id)
                hog.active = False
                self._frozen_hogs_before.add(hog)
        for h in hogids_to_remove:
            self.subhogs.pop(h)
        self._rep_lookup = self._prepare_lookups()


    def infer_genetree_from_msa(self, msa):
        genetree_nwk = _wrappers.infer_gene_tree(msa)
        try:
            genetree = Tree(genetree_nwk + ";", format=0, quoted_node_names=True)
        except ValueError:
            logger.error(f"cannot load genetree from {genetree_nwk}")
            raise RuntimeError("cannot load genetree from {}".format(genetree_nwk))
        self.write_msa_or_tree_if_necessary(genetree, fn_suffix=".nwk")
        return genetree

    def infer_rooted_genetree(self, gene_tree: TreeNode):
        genetree = deepcopy(gene_tree)
        if self.conf.gene_rooting_method == "midpoint":
            r_outgroup = genetree.get_midpoint_outgroup()
            genetree.set_outgroup(r_outgroup)  # print("Midpoint rooting is done for gene tree.")

        elif self.conf.gene_rooting_method == "midpoint-dendropy":
            dt = dendropy.Tree.get_from_string(gene_tree.write(format=1), schema="newick")
            dt.reroot_at_midpoint()
            nw = dt.as_string(schema="newick", suppress_rooting=True)
            genetree = Tree(nw, format=1, quoted_node_names=True)

        elif self.conf.gene_rooting_method == "Nevers_rooting":
            logger.info("Nevers_rooting started for " + str(genetree.write(format=1, format_root_node=True)))
            species = Tree("species_tree.nwk", format=1)
            genetree = _utils_subhog.get_score_all_root(genetree, species)
            logger.info("Nevers_rooting finished for " + str(genetree.write(format=1, format_root_node=True)))

        elif self.conf.gene_rooting_method == "mad":
            genetree = _wrappers.mad_rooting(genetree)  # todo check with qouted gene tree

        elif self.conf.gene_rooting_method == "outlier":  # not yet implmented completely, todo need check with new gene tree
            outliers = _utils_subhog.find_outlier_leaves(genetree)
            r_outgroup = _utils_subhog.midpoint_rooting_outgroup(genetree, leaves_to_exclude=outliers)
            genetree.set_outgroup(r_outgroup)
        else:
            logger.warning("rooting method not found !!   * * * * *  *")
            raise ValueError("invalid rooting method: {}".format(self.conf.gene_rooting_method))
        return genetree

    def infer_reconciliation(self, genetree:TreeNode, sos_threshold=0.0):
        """Annotate each internal node with a 'evoltype' and 'sos' attribute.

        The evoltype attribute will be either 'S' or 'D' (speciation or duplication),
        and the 'sos' attribute will contain the fraction of species that overlap among the
        subtrees. The parameter sos_threshold specifies the minimum fraction of species that
        must overlap among the subtrees in order to be considered a duplication.

        :param genetree: the rooted tree that should be processed
        :param sos_threshold: the minimum fraction of species that must overlap among the
                              subtrees in order to be considered a duplication.
        """
        logger.info("\n" + genetree.get_ascii(show_internal=True, attributes=['name', 'hogid']))
        cnt_D, cnt_S = 0, 0
        for n in genetree.traverse('postorder'):
            if n.is_leaf():
                n.add_feature('species', self._rep_lookup[n.name].representative.get_species())
                n.add_feature('hogid', self._rep_lookup[n.name].hog.hogid)
            else:
                assert len(n.children) >= 2, f"{n.name} has {len(n.children)} children"
                specs = tuple(c.species for c in n.children)
                for c in n.children:
                    c.del_feature('species')  # cleanup to avoid excessive memory
                sp_inter = set.intersection(*specs)
                sp_union = set.union(*specs)
                sos = len(sp_inter) / len(sp_union)
                n.add_feature('species', sp_union)
                n.add_feature('sos', sos)
                n.add_feature('so_tuple', (len(sp_inter), len(sp_union)))
                n.add_feature('evoltype', 'D' if sos > sos_threshold else 'S')
                if sos > sos_threshold:
                    cnt_D += 1
                    n.name = f"D{cnt_D}_{len(sp_inter)}_{len(sp_union)}"
                else:
                    cnt_S += 1
                    n.name = f"S{cnt_S}"
        genetree.del_feature('species')
        self.write_msa_or_tree_if_necessary(genetree, fn_suffix="_rec.nwk", features=['evoltype', 'sos', 'so_tuple', 'hogid'])

    def _compute_between_and_within_distances(self, genetree, subtrees_node, partitions):
        mrca = genetree.get_common_ancestor(*subtrees_node)
        id2part = {id_: i for i, p in enumerate(partitions) for id_ in p}
        N = len(partitions)
        dists = collections.defaultdict(list)
        for l1, l2 in itertools.combinations(mrca.iter_leaves(), 2):
            i1 = id2part.get(l1.name, N)
            i2 = id2part.get(l2.name, N)
            dists[(i1, i2)].append(mrca.get_distance(l1, l2))

        return dists

    def _resolve_conflicts(self, gene_tree: TreeNode, subtrees_node, partitions) -> bool:
        """resolves cases of conflicting subhog splittings where
        the representatives of a subhog are split among different subtrees
        and have been merged even before the previous level.

        the method returns a boolean value indicating whether a new reconciliation
        is necessary (e.g. in case the tree has been modified and hence the labeling
        of speciation/duplication might have changed).
        """
        updated_subtree_nodes = []
        for n in subtrees_node:
            # previous rounds of _resolve_conflicts might have left disconnected subtree_nodes
            # so we identify the first node which is either a leaf or a bifurcating node.
            while len(n.children) == 1:
                n = n.children[0]
            updated_subtree_nodes.append(n)
        subtrees_node = updated_subtree_nodes
        mrca = gene_tree.get_common_ancestor(*subtrees_node)
        min_support = min(n.support for n in subtrees_node)
        min_branch_to_subtree = min(gene_tree.get_distance(mrca, n) for n in subtrees_node)
        partitions = sorted(partitions, key=lambda x: [-sum(len(self._rep_lookup[r].representative.get_subelements()) for r in x), -len(x),])
        if False and (min_support < 0.7 or min_branch_to_subtree < 0.01):
            # we collapse things, i.e. separate nodes in tree based on hogid
            # we mv all leaves from the non-biggest partitions to the mrca of the biggest partion
            if len(partitions[0]) > 1:
                target_node = gene_tree.get_common_ancestor(partitions[0])
            else:
                target_node = gene_tree.search_nodes(name=partitions[0][0])[0].up
            for small_partition in partitions[1:]:
                for name in small_partition:
                    n = gene_tree.search_nodes(name=name)[0]
                    parent = n.up
                    n = n.detach()
                    target_node.add_child(n)
                    if len(parent.children) == 1 and parent.up is None:
                        parent.children[0].delete(prevent_nondicotomic=True, preserve_branch_length=True)
                    elif len(parent.children) == 1:
                        parent.delete(prevent_nondicotomic=True, preserve_branch_length=True)
                    # deactivate representative for next level
                    self._rep_lookup[n.name].representative.enabled = False
            return True
        else:
            # we remove the bogus representatives, but keep the labeling of the
            # speciation/duplication nodes. that way, the two subhogs will not get merged
            # at this level (since they are in different subtrees).
            stats = [{'len': len(p),
                      'rep_size': sum(len(self._rep_lookup[r].representative.get_subelements()) for r in p),
                     } for p in partitions]
            logger.info(f"Resolving conflict for {self._rep_lookup[partitions[0][0]].hog.hogid} by ignoring representatives {partitions[1:]}")
            logger.debug(f"{len(partitions)} partitions: {stats}")
            for small_partition in partitions[1:]:
                for name in small_partition:
                    n = gene_tree.search_nodes(name=name)[0]
                    n.delete(prevent_nondicotomic=True, preserve_branch_length=True)
            return False

    def merge_subhogs(self, reconciled_genetree:TreeNode, msa:MultipleSeqAlignment):
        while True:
            hogids_in_subtrees = collections.defaultdict(list)
            for n in self._get_merge_candidates_with_hogids(reconciled_genetree):
                for hogid in n.hogids:
                    hogids_in_subtrees[hogid].append(n)
            if any(len(z) > 1 for z in hogids_in_subtrees.values()):
                logger.info("At least one subhog is split. here is the full labeled genetree:\n"
                            + reconciled_genetree.get_ascii(show_internal=True, attributes=['evoltype', 'sos', 'hogids', 'hogid']))
            else:
                break
            redo_reconcilation = False
            for hogid, subtrees in hogids_in_subtrees.items():
                if len(subtrees) > 1:
                    logger.info(f"Representaives of {hogid} are split among {len(subtrees)} candidate subtrees.")
                    split_parts = [list(n.name for n in sub.iter_leaves() if n.hogid == hogid) for sub in subtrees]
                    split_hogs = split_hog(self.subhogs[hogid], *split_parts)
                    if split_hogs and len(split_hogs) > 1:
                        # we could split the current hog.
                        self.subhogs.pop(hogid)
                        self.subhogs.update({h.hogid: h for h in split_hogs})
                    else:
                        redo_reconcilation |= self._resolve_conflicts(reconciled_genetree, subtrees, split_parts)
            self._rep_lookup = self._prepare_lookups()
            if redo_reconcilation:
                if len(reconciled_genetree.children) == 1:
                    reconciled_genetree.children[0].delete(preserve_branch_length=True)
                self.infer_reconciliation(reconciled_genetree, sos_threshold=self.conf.threshold_dubious_sd)
        new_hogs = []
        processed_nodes = set([])
        for subtrees in hogids_in_subtrees.values():
            for subtree in subtrees:
                if subtree in processed_nodes:
                    continue
                processed_nodes.add(subtree)
                new_repr = self.find_most_divergent_representatives_from_genetree(subtree)
                hog = HOG([self.subhogs[h] for h in subtree.hogids],
                          self.node_species_tree,
                          rhogid=self.rhogid,
                          msa=msa,
                          representatives=new_repr,
                          conf_infer_subhhogs=self.conf)
                new_hogs.append(hog)
        return new_hogs

    def process(self) -> List[HOG]:
        """Process the given roothog at the given taxonomic level.

        This method does all the processessing of the subhogs from the
        previous taxonomic levels. It does

        1. Align the subhogs from the previous levels as a new MSA.
        2. Filter the initial MSA (remove gappy columns and rows)
        3. Infer a genetree for the new filtered MSA
        4. Root (mid-point) and reconcile (with species-overlap) the gene tree
        5. Merge the subhogs from the previous levels into new HOGs
        """
        msa = self.align_subhogs()
        filtered_msa = self.filter_msa(msa)
        if len(filtered_msa) == 0:
            return list(self.subhogs.values())
        gene_tree = self.infer_genetree_from_msa(filtered_msa)
        if len(gene_tree) > 1:
            rooted_gene_tree = self.infer_rooted_genetree(gene_tree)
            # TODO: dealing with fragments is not done yet.
            self.infer_reconciliation(rooted_gene_tree, sos_threshold=self.conf.threshold_dubious_sd)
        else:
            rooted_gene_tree = gene_tree
        new_hogs = self.merge_subhogs(rooted_gene_tree, msa=filtered_msa)
        new_hogs.extend(self._frozen_hogs_before)
        return new_hogs


def infer_hogs_this_level(node_species_tree, rhogid, pickles_subhog_folder_all, conf_infer_subhhogs):
    """
    infer subHOGs for a rootHOG at a taxanomic level
    """

    this_level_node_name = node_species_tree.name
    pickles_subhog_folder = pickles_subhog_folder_all + "/rhog_" + rhogid + "/"
    if node_species_tree.is_leaf():
        logger.warning("issue 1235,it seems that there is only on species in this tree, singleton hogs are treated elsewhere "+ rhogid)

    pickle_subhog_file = pickles_subhog_folder + str(this_level_node_name) + ".pickle"

    # TODO arrage resume with nextflow and also for when read single_tone pickles
    if inferhog_resume_subhog:
        if os.path.exists(pickle_subhog_file) and os.path.getsize(pickle_subhog_file) > 3:  # 3 bytes
            with open(pickle_subhog_file, 'rb') as handle:
                # todo : do I really need to read the pickle file
                hogs_this_level_list = pickle.load(handle)  #[object class HOG HOG:4027_sub1,len=1,taxono=PSETE]
                if hogs_this_level_list:
                    logger.debug("Level " + str(this_level_node_name) + " with " + str(len(hogs_this_level_list)) + " hogs is read from pickle.")
                    return len(hogs_this_level_list)
                else:
                    logger.debug(" Issue  1238510: the pickle file for single tone is empty " + str(hogs_this_level_list) + " " + rhogid)

    hogs_children_level_list = read_children_hogs(node_species_tree, rhogid, pickles_subhog_folder_all)

    if len(hogs_children_level_list) == 1:
        hogs_this_level_list = hogs_children_level_list
        with open(pickle_subhog_file, 'wb') as handle:
            pickle.dump(hogs_this_level_list, handle, protocol=pickle.HIGHEST_PROTOCOL)
        return len(hogs_children_level_list)

    level_processor = LevelHOGProcessor(node_species_tree, hogs_children_level_list, rhogid, conf_infer_subhhogs)
    hogs_this_level_list = level_processor.process()

    with open(pickle_subhog_file, 'wb') as handle:
        pickle.dump(hogs_this_level_list, handle, protocol=pickle.HIGHEST_PROTOCOL)
    return len(hogs_this_level_list)


def merge_subhogs(gene_tree, hogs_children_level_list, node_species_tree, rhogid, merged_msa, conf_infer_subhhogs):
    """
    merge subhogs based on the gene tree specieciaton node of gene tree by creating inter-HOG graph (implicitley )
    """

    subhogs_id_children_assigned = []  # the same as  subHOG_to_be_merged_all_id
    hogs_this_level_list = []
    subHOG_to_be_merged_set_other_Snodes = []
    subHOG_to_be_merged_set_other_Snodes_flattned_temp = []
    ##  the following if for debugging and visualisation of connected component of inter-HOG graph
    # hoggraph_node_name = [i._hogid.split("_")[1][3:] for i in hogs_children_level_list]
    # hog_size_dic = {}
    # dic_hog = {}
    # for hog in hogs_children_level_list:
    #     hog_id_short = hog._hogid.split("_")[1][3:]
    #     for prot in hog._members:
    #         dic_hog[prot] = hog_id_short + "_len" + str(len(hog._members))
    #     hog_size_dic[hog_id_short] = len(hog._members)
    # for hog in hogs_children_level_list:
    #     hog_id_short = hog._hogid.split("_")[1][3:] + "_len" + str(len(hog._members))
    #     # print(hog_id_short, ":", hog._members)
    # hoggraph_node_name_len = [i+"_len"+str(hog_size_dic[i]) for i in hoggraph_node_name]
    # hoggraph = nx.Graph()
    # hoggraph.add_nodes_from(hoggraph_node_name_len)

    for node in gene_tree.traverse(strategy="preorder", is_leaf_fn=lambda n: hasattr(n, "processed") and n.processed==True):

        if not node.is_leaf() and node.name[0] == "S":
            node_leaves_name_raw0 = [i.name for i in node.get_leaves()]
            # leaves names  with subhog id  'HALSEN_R15425||HALSEN|_|1352015793_sub10149',
            node_leaves_name_raw1 = [i.split("|_|")[0] for i in node_leaves_name_raw0]

            if node_leaves_name_raw1[0].startswith("'"):
                node_leaves_name = [i[1:] for i in node_leaves_name_raw1] # since splited , there is no ' at the end. so -1] is not needed.
                # "'sp|O67547|SUCD_AQUAE||AQUAE||1002000005", gene tree is quoted, there are both ' and " !
            else:
                node_leaves_name = node_leaves_name_raw1
            # node_leaves_name =[]
            # for name_i in node_leaves_name_:
            #     node_leaves_name += name_i.split("_|_")

            # num_prot = len(s_gene_tree_leaves)
            # for i in range(num_prot):
            #     hog_i = dic_hog[s_gene_tree_leaves[i]]
            #     for j in range(i):
            #         hog_j = dic_hog[s_gene_tree_leaves[j]]
            #         if hoggraph.has_edge(hog_i, hog_j):
            #             hoggraph[hog_i][hog_j]['weight'] += 1
            #         else:
            #             hoggraph.add_edge(hog_i, hog_j, weight=1)
            # print(hoggraph.edges(data=True))
            # todo this piece of code could be neater, for extracting connected component of inter-hog graph
            subHOG_to_be_merged = []
            for node_leave_name in node_leaves_name:  # print(node_leave_name)
                for subHOG in hogs_children_level_list:
                    subHOG_members = subHOG._members
                    if node_leave_name in subHOG_members:  # could be improved
                        if subHOG._hogid not in subHOG_to_be_merged_set_other_Snodes_flattned_temp: #todo 18march this means that we are not merging a few together
                            subHOG_to_be_merged.append(subHOG)
                            subhogs_id_children_assigned.append(subHOG._hogid)
                        else:  # this hog is already decided to be merged  print(node.name, subHOG._hogid, node_leave_name)
                            if "processed" in node:
                                logger.warning("issue 1863 "+ str(node.name)+str(subHOG._hogid)+ str(node_leave_name)) # print("processed", node.name) #else: #    print("processed not in ", node.name)  # print(node_leave_name,"is in ",subHOG._hogid)
            if len(subHOG_to_be_merged) == 1:
                logger.warning("issue 125568313 "+str(subHOG_to_be_merged)+" "+node.name+" probably the subhog was merged previously" )
                logger.debug("issue 125568313 " + str(subHOG_to_be_merged_set_other_Snodes_flattned_temp))
                logger.debug("issue 125568313 " +str(node.name)+" "+ str(gene_tree.write(format=1,format_root_node=True)))

            elif len(subHOG_to_be_merged)>1:
                subHOG_to_be_merged_set = set(subHOG_to_be_merged)
                HOG_this_node = HOG(subHOG_to_be_merged_set, node_species_tree, rhogid, merged_msa, conf_infer_subhhogs)
                if len(HOG_this_node._msa) == 1:
                    logger.warning("issue 1258313"+str(HOG_this_node)+str(HOG_this_node._msa)+" "+node.name  )
                hogs_this_level_list.append(HOG_this_node)

                subHOG_to_be_merged_set_other_Snodes.append([i._hogid for i in subHOG_to_be_merged_set])
                subHOG_to_be_merged_set_other_Snodes_flattned_temp = [item for items in subHOG_to_be_merged_set_other_Snodes for  item in items]
                #  I don't need to traverse deeper in this clade
            node.processed = True  # print("?*?*  ", node.name)
        subHOG_to_be_merged_set_other_Snodes_flattned = [item for items in subHOG_to_be_merged_set_other_Snodes for item in items]
        if [i._hogid for i in hogs_children_level_list] == subHOG_to_be_merged_set_other_Snodes_flattned:
            break
    ##  the following if fore debugging and visualisation of connected component of inter-HOG graph
    # fig = plt.figure(figsize=(300, 200), dpi=60)
    # pos = nx.spring_layout(hoggraph, k=0.25, iterations=30)  # For better example looking  # smaller k, biger space between
    # nx.draw(hoggraph, pos, with_labels=True, node_color='y', node_size=500, font_size=16) # , alpha=0.4
    # # nx.draw(G, pos,, edge_color="r", font_size=16, with_labels=True)
    # labels = {e: hoggraph.edges[e]['weight'] for e in hoggraph.edges}
    # nx.draw_networkx_edge_labels(hoggraph, pos, edge_labels=labels, font_size=16)
    # num = random.randint(3, 1000000)
    # plt.savefig("./hoggraph/" + hogs_children_level_list[0]._hogid[4:] + "file_rndm"+str(num)+".jpg")
    # plt.show()
    for subHOG in hogs_children_level_list:  # for the single branch  ( D include subhog and S node. )
        if subHOG._hogid not in subhogs_id_children_assigned:  # print("here", subHOG)
            hogs_this_level_list.append(subHOG)
    # if len(hogs_this_level_list)==1:  hogs_this_level_list = [hogs_this_level_list]

    for hog_j in hogs_this_level_list:
        hog_j._tax_now = node_species_tree
    ##  the following if fore debugging of connected component of inter-HOG graph
    ## check for conflicts in merging
    #     for i in range(subHOG_to_be_merged_set_other_Snodes):  if
    #         for i in range(subHOG_to_be_merged_set_other_Snodes):  print("*&*& ",node_species_tree.name)
    # prot_list_sbuhog = [i._members for i in hogs_this_level_list]
    # prot_list_sbuhog_short = []
    # for prot_sub_list_sbuhog in prot_list_sbuhog:
    #     if format_prot_name == 0:  # bird dataset TYTALB_R04643
    #         prot_list_sbuhog_short = prot_sub_list_sbuhog
    #     elif format_prot_name == 1:  # qfo dataset  'tr|E3JPS4|E3JPS4_PUCGT
    #         prot_list_sbuhog_short.append([prot.split('|')[2] for prot in prot_sub_list_sbuhog])
    # logger_hog.debug(str(len(hogs_this_level_list))+" hogs are inferred at the level "+node_species_tree.name+": "+' '.join(
    #     [str(i) for i in prot_list_sbuhog_short]))

    return hogs_this_level_list



from .filter_orthoxml import *
from .convert import orthoxml_to_newick
from .orthoxml_merge import merge_orthoxml_files
from .extract_groups import TaxLevel, extract_flat_groups_at_level, extract_marker_groups_at_level



import os, types, itertools
from abc import ABCMeta, abstractmethod
from enum import Enum
from Bio import AlignIO, SeqIO
from Bio.Align import MultipleSeqAlignment
from ...seq_utils import is_dna

from FastOMA.zoo.wrappers import WrapperError
from FastOMA.zoo.wrappers.aligners.base_aligner import identify_input

import logging
logger = logging.getLogger(__name__)

AlignmentInput = Enum('AlignmentInput', 'OBJECT FILENAME')
DataType = Enum('DataType', 'DNA PROTEIN UNKNOWN')


class TreeBuilder(object):
    """
    Base class for wrappers of tree building software

    The wrapper is written as a callable class.
    This can hold data (state) to do with the operation it performs, so it can keep results,
    execution times and other metadata, as well as perform the task.

    This is a base implementation to be extended. The important parts are
    __init__ (does the setup) and __call__ (does the work). All
    else are helper methods.

    :Example:

    ::

        callable_wrapper = ConcreteAligner(aln)
        result = callable_wrapper()
        time_taken = callable_wrapper.elapsed_time
        result_again = callable_wrapper.result
    """
    __metaclass__ = ABCMeta

    def __init__(self, alignment=None, datatype=DataType.UNKNOWN, binary=None):
        """
        ..note::  TODO: this documentation is not correct. it needs to be updateted.

        Should work the same whether you're working with a Biopython object or a file
            but the implementation differs, e.g. a Biopython object will need
            to be written temporarily to disk for the Aligner to work on it.

        alignment is one of 4 things:
            a filename
            a Biopython MSA
            a list of Seq objects
            anything else (throw an exception)

        binary is the alignment's executable file, or None
        """

        if alignment is not None:
            self.input_type = identify_input(alignment)  # Figure out what it is - file or object
            if datatype == DataType.UNKNOWN:
                # dup, input_ = itertools.tee(input_)
                self.datatype = guess_datatype(alignment, from_filename=self.input_type == AlignmentInput.FILENAME)
            else:
                self.datatype = datatype

            self.input = alignment  # store it
        else:
            self.input_type = None
            self.input = None

        self.elapsed_time = None
        self.stdout = None
        self.stderr = None
        try:
            self.cli = self._init_cli(binary)
        except IOError as err:
            raise WrapperError('Error searching for binary: {}'.format(err))
            # End setup

    @abstractmethod
    def __call__(self, *args, **kwargs):
        """
        How to call the underlying aligner
        """
        pass

    @abstractmethod
    def _init_cli(self, binary):
        """
        Set up the command-line interface to the wrapped software
        :param binary: filename of executable binary file
        :return: concrete CLI type inheriting from AbstractCLI
        """
        pass


def guess_datatype(alignment, from_filename=False):
    if from_filename:
        try:
            alignment = list(SeqIO.parse(alignment, 'fasta'))
        except:
            alignment = list(SeqIO.parse(alignment, 'phylip-relaxed'))
    if is_dna(alignment):
        logger.warning("Guessed datatype=DNA. But better explicitly specify the sequence type with option datatype={DNA, PROTEIN}.")
        return DataType.DNA 
    else:
        logger.warning("Guessed datatype=PROTEIN. But better explicitly specify the sequence type with option datatype={DNA, PROTEIN}.")
        return DataType.PROTEIN


def identify_input(alignment):
    """
    Work out if we're dealing with an alignment (return True), a file
    (return False), or invalid input (raise error)

    :param alignment: either an Biopython MultipleSequenceAlignment or
        a filename pointing to an existing msa file.
    """
    try:
        if isinstance(alignment, (MultipleSeqAlignment, types.GeneratorType, list)):
            # `alignment` is a Biopython MultipleSequenceAlignment
            return AlignmentInput.OBJECT

        elif isinstance(alignment, str) and os.path.exists(alignment):
            # `alignment` is a filepath
            return AlignmentInput.FILENAME

    except:
        # `alignment` is some other thing we can't handle
        raise ValueError('{} is not an alignment object or a valid filename'.format(alignment))


